@@ -1,87 +1,35 @@
-"""Vectorized GLM model with sk-learn type API"""
+"""Generalized Linear Model with prior control for regression in scikit-learn type API."""
 
 import warnings
-from pathlib import Path
-from typing import Any, Dict, List, Optional, Union
+from typing import Any, Dict, List, Optional, Tuple, Union
 
 import numpy as np
+import scipy.sparse as sp
 import scipy.stats as stats
-from cmdstanpy import CmdStanModel, set_cmdstan_path
 from numpy.typing import ArrayLike, NDArray
 
 from scikit_stan.modelcore import CoreEstimator
+from scikit_stan.utils.stan import load_stan_model
 from scikit_stan.utils.validation import (
     FAMILY_LINKS_MAP,
+    GLM_FAMILIES,
     check_array,
     check_is_fitted,
+    method_dict,
     validate_aux_prior,
     validate_family,
     validate_prior,
 )
 
-STAN_FILES_FOLDER = Path(__file__).parent.parent / "stan_files"
-CMDSTAN_VERSION = "2.30.1"
-
-method_dict = {
-    "sample": CmdStanModel.sample,
-    "optimize": CmdStanModel.optimize,
-    "variational": CmdStanModel.variational,
-}
-
-GLM_FAMILIES = {
-    "gaussian": 0,
-    "gamma": 1,
-    "inverse-gaussian": 2,
-    "poisson": 3,
-    "binomial": 4,
-    "negative-binomial": 5,
-}
-
-# handle pre-compiled models and possibly repackaged cmdstan
-
-local_cmdstan = STAN_FILES_FOLDER / f"cmdstan-{CMDSTAN_VERSION}"
-if local_cmdstan.exists():
-    set_cmdstan_path(str(local_cmdstan.resolve()))
-
-try:
-    GLM_CONTINUOUS_STAN = CmdStanModel(
-        exe_file=STAN_FILES_FOLDER / "glm_v_continuous.exe",
-        stan_file=STAN_FILES_FOLDER / "glm_v_continuous.stan",
-        compile=False,
-    )
-
-    GLM_DISCRETE_STAN = CmdStanModel(
-        exe_file=STAN_FILES_FOLDER / "glm_v_discrete.exe",
-        stan_file=STAN_FILES_FOLDER / "glm_v_discrete.stan",
-        compile=False,
-    )
-except ValueError:
-    import shutil
-
-    warnings.warn("Failed to load pre-built models, compiling")
-    GLM_CONTINUOUS_STAN = CmdStanModel(
-        stan_file=STAN_FILES_FOLDER / "glm_v_continuous.stan",
-        stanc_options={"O1": True},
-    )
-    GLM_DISCRETE_STAN = CmdStanModel(
-        stan_file=STAN_FILES_FOLDER / "glm_v_discrete.stan",
-        stanc_options={"O1": True},
-    )
-    shutil.copy(
-        GLM_CONTINUOUS_STAN.exe_file,  # type: ignore
-        STAN_FILES_FOLDER / "glm_v_continuous.exe",
-    )
-    shutil.copy(
-        GLM_DISCRETE_STAN.exe_file,  # type: ignore
-        STAN_FILES_FOLDER / "glm_v_discrete.exe",
-    )
+GLM_CONTINUOUS_STAN = load_stan_model("glm_v_continuous")
+GLM_DISCRETE_STAN = load_stan_model("glm_v_discrete")
 
 
 class GLM(CoreEstimator):
     r"""
     A generalized linear model estimator with several options for families, links,
     and priors on regression coefficients, the intercept, and error scale,
-    done in an sk-learn style.
+    done in an scikit-learn style.
     This class also provides an autoscaling feature of the priors.
     For deterministic behavior from this model, the class's seed can be set and is then
     passed to Stan computations.
@@ -102,7 +50,7 @@ class GLM(CoreEstimator):
         this documentation for more information:
         https://mc-stan.org/cmdstanpy/api.html#cmdstanmodel
 
-        Customizing these fields occurs as a passed dictionary, which is validated
+        Customizing these fields is done via a passed dictionary, which is validated
         on the level of CmdStan. As an example, to specify the number of chains for
         the HMC-NUTS sampler to run, it is sufficient to pass::
 
@@ -120,6 +68,14 @@ class GLM(CoreEstimator):
 
         Default Stan parameters are used if nothing is passed.
 
+    fit_intercept : bool, optional
+        Whether the GLM should fit a global intercept. Some hierarchical models operate
+        without an intercept, so setting this to False will fill the first column of the
+        predictor matrix with ones.
+
+        By default, the GLM has an intercept term -- the first column of the predictor
+        matrix X is used for inference of the intercept.
+
     family : str, optional
         Distribution family used for linear regression. All R Families package are supported:
 
@@ -174,8 +130,11 @@ class GLM(CoreEstimator):
         via :class:`numpy.random.RandomState`.
         Specifying this field will yield the same result for multiple uses if
         all other parameters are held the same.
+
     priors : Optional[Dict[str, Union[int, float, List]]], optional
         Dictionary for configuring prior distribution on coefficients.
+        Currently supported priors are: "normal", and "laplace".
+
         By default, all regression coefficient priors are set to
 
         .. math:: \beta \sim \text{normal}(0, 2.5 \cdot \text{sd}(y) / \text{sd}(X))
@@ -215,6 +174,10 @@ class GLM(CoreEstimator):
             }
 
         Any unspecified priors will be set to the default.
+
+        Also note that choosing a Laplace prior is equivalent to L1 regularization:
+        https://stats.stackexchange.com/questions/177210/why-is-laplace-prior-producing-sparse-solutions/177217#177217
+
     prior_intercept : Optional[Dict[str, Any]], optional
         Prior for the intercept alpha parameter for GLM.
         If this is not specified, the default is
@@ -248,11 +211,16 @@ class GLM(CoreEstimator):
 
         by default (without autoscaling, see below).
 
+        Also note that choosing a Laplace prior is equivalent to L1 regularization:
+        https://stats.stackexchange.com/questions/177210/why-is-laplace-prior-producing-sparse-solutions/177217#177217
+
     prior_aux : Optional[Dict[str, Any]], optional
         Prior on the auxiliary parameter for the family used in
         the regression: for example, the std for Gaussian, shape for gamma, etc...
-        Currently supported priors are: "exponential" and "chi2", which are
-        both parameterized by a single scalar.
+        Currently supported priors are: "exponential", "chi2", "gamma", and "inv_gamma".
+        These are parameterized by the following distribution parameters:
+        "beta", "nu", "alpha" and "beta", and "alpha" and "beta", respectively.
+
         Priors here with more parameters are a future feature.
 
         If an empty dictionary {} is passed, the default Stan uniform(-inf, inf) prior is used.
@@ -261,13 +229,14 @@ class GLM(CoreEstimator):
 
             + "prior_aux_dist": distribution of the prior on this parameter
             + "prior_aux_param": parameter of the prior on this parameter
+                                - see above for admissible tags
 
         For example, to specify a chi2 prior with nu=2.5, pass::
 
             {
                 "prior_aux_dist": "chi2",
 
-                "prior_aux_param": 2.5
+                "nu": 2.5
             }
 
         The default un-scaled prior is ``exponential(1)``, the default scaled prior is
@@ -292,6 +261,7 @@ class GLM(CoreEstimator):
         self,
         algorithm: str = "sample",
         algorithm_params: Optional[Dict[str, Any]] = None,
+        fit_intercept: bool = True,
         family: str = "gaussian",
         link: Optional[str] = None,
         seed: Optional[int] = None,
@@ -303,6 +273,8 @@ class GLM(CoreEstimator):
         self.algorithm = algorithm
         self.algorithm_params = algorithm_params
 
+        self.fit_intercept = fit_intercept
+
         self.family = family
         self.link = link
 
@@ -320,7 +292,7 @@ class GLM(CoreEstimator):
         show_console: bool = False,
     ) -> "CoreEstimator":
         """
-        Fits GLM object to the given data.
+        Fits GLM model to the given data.
         This model is considered fit once its alpha, beta,
         and sigma parameters are determined via a regression.
 
@@ -329,6 +301,7 @@ class GLM(CoreEstimator):
         X : ArrayLike
             NxK matrix of predictors, where K >= 0. If K = 1,
             then X is automatically reshaped to being 2D and raises a warning.
+            X can be sparse. It will be converted to CSR format if it is not already.
         y : ArrayLike
             Nx1 outcome vector where each row is a response corresponding to
             the same row of predictors in X.
@@ -409,6 +382,7 @@ class GLM(CoreEstimator):
 
         validate_family(self.family, self.link_)
 
+        # ensure that identity link function is not used for discrete regressions
         if not self.is_cont_dat_ and self.link_ == "identity":
             self.link_ = "logit" if self.family == "bernoulli" else "log"
 
@@ -420,13 +394,13 @@ class GLM(CoreEstimator):
 
         # data common to all prior choices for intercept and coefficients
         dat = {
-            "X": X_clean,
             "y": y_clean,
             "N": X_clean.shape[0],
             "K": K,
             "family": self.familyid_,
             "link": self.linkid_,
             "predictor": 0,
+            "fit_intercept": self.fit_intercept,
             "prior_intercept_dist": None,
             "prior_intercept_mu": None,
             "prior_intercept_sigma": None,
@@ -435,7 +409,13 @@ class GLM(CoreEstimator):
             "prior_slope_sigma": [None] * K,
         }
 
-        sdx = np.std(X_clean)
+        prepare_X_data(dat, X_clean)
+
+        if sp.issparse(X_clean):
+            # .std does not work on sparse matrices, so do this instead
+            sdx = np.sqrt((X_clean.power(2)).mean() - np.square(X_clean.mean()))
+        else:
+            sdx = X_clean.std()
         if self.familyid_ == 0:  # gaussian
             my = np.mean(y_clean) if self.linkid_ == 0 else 0.0
             sdy = np.std(y_clean)
@@ -446,8 +426,6 @@ class GLM(CoreEstimator):
         if sdy == 0.0:
             sdy = 1.0
 
-        dat["sdy"] = sdy
-
         DEFAULT_SLOPE_PRIOR = {
             "prior_slope_dist": 0,
             "prior_slope_mu": [0.0] * K,
@@ -529,6 +507,7 @@ class GLM(CoreEstimator):
         # validate auxiliary parameter prior
         if self.prior_aux is None:
             self.prior_aux_ = {
+                "num_prior_aux_params": 1,
                 "prior_aux_dist": 0,  # exponential
             }
 
@@ -538,7 +517,7 @@ class GLM(CoreEstimator):
                         sigma ~ exponential(1 / sd(y))
                     """
                 )
-                self.prior_aux_["prior_aux_param"] = 1.0 / sdy
+                self.prior_aux_["prior_aux_params"] = [1.0 / sdy]
             else:
                 warnings.warn(
                     """Prior on auxiliary parameter not specified. Using default unscaled prior
@@ -546,14 +525,15 @@ class GLM(CoreEstimator):
                     """
                 )
 
-                self.prior_aux_["prior_aux_param"] = 1.0
+                self.prior_aux_["prior_aux_params"] = [1.0]
         else:
             # set auxiliary parameter prior to be default flat as in Stan:
             # uniform(-infinity, +infinity)
             if len(self.prior_aux) == 0:
                 self.prior_aux_ = {
+                    "num_prior_aux_params": 1,
                     "prior_aux_dist": -1,
-                    "prior_aux_param": 0.0,
+                    "prior_aux_params": [0.0],
                 }
             else:
                 self.prior_aux_ = validate_aux_prior(self.prior_aux)
@@ -568,12 +548,16 @@ class GLM(CoreEstimator):
         dat["prior_slope_sigma"] = self.priors_["prior_slope_sigma"]
 
         dat["prior_aux_dist"] = self.prior_aux_["prior_aux_dist"]
-        dat["prior_aux_param"] = self.prior_aux_["prior_aux_param"]
+        dat["num_prior_aux_params"] = self.prior_aux_["num_prior_aux_params"]
+        dat["prior_aux_params"] = self.prior_aux_["prior_aux_params"]
 
         if self.is_cont_dat_:
             self.model_ = GLM_CONTINUOUS_STAN
         else:
             self.model_ = GLM_DISCRETE_STAN
+            # XXX this seems like mostly nonsense?
+            # needs to be separate keyword argument or make Y 2d?
+            # binomial should be its own model class
             dat["trials"] = np.repeat(y_clean.shape[0], X_clean.shape[0])
 
         self.seed_ = self.seed
@@ -592,8 +576,9 @@ class GLM(CoreEstimator):
 
         stan_vars = self.fitted_samples_.stan_variables()
         if self.algorithm == "sample":
-            self.alpha_ = stan_vars["alpha"].mean(axis=0)
-            self.alpha_samples_ = stan_vars["alpha"]
+            if self.fit_intercept:
+                self.alpha_ = stan_vars["alpha"].mean(axis=0)
+                self.alpha_samples_ = stan_vars["alpha"]
 
             self.beta_ = stan_vars["beta"].mean(axis=0)
             self.beta_samples_ = stan_vars["beta"]
@@ -603,7 +588,9 @@ class GLM(CoreEstimator):
                 self.sigma_ = stan_vars["sigma"].mean(axis=0)
                 self.sigma_samples_ = stan_vars["sigma"]
         else:
-            self.alpha_ = stan_vars["alpha"]
+            if self.fit_intercept:
+                self.alpha_ = stan_vars["alpha"]
+
             self.beta_ = stan_vars["beta"]
 
             if self.is_cont_dat_:
@@ -627,6 +614,7 @@ class GLM(CoreEstimator):
         ----------
         X : NDArray[Union[np.float64, np.int64]]
             Predictor matrix or array of data to use for prediction.
+            X can be sparse. It will be converted to CSR format if it is not already.
         show_console : bool, optional
             Printing output of default CmdStanPy console during Stan operations.
 
@@ -674,15 +662,16 @@ class GLM(CoreEstimator):
                     random_state=self.seed_,
                 )
 
-        dat = {
+        dat: Dict[str, Any] = {
             "N": X_clean.shape[0],
             "K": X_clean.shape[1],
-            "X": X_clean,
             "y": [],
+            # XXX: setting empty breaks discrete GQ
             "trials": [],
             "family": self.familyid_,
             "link": self.linkid_,
             "predictor": 1,
+            "fit_intercept": self.fit_intercept,
             "prior_intercept_dist": 0,
             "prior_intercept_mu": 1.0,
             "prior_intercept_sigma": 2.5,
@@ -691,9 +680,13 @@ class GLM(CoreEstimator):
             "prior_slope_sigma": [0.0] * X_clean.shape[1],
             "prior_aux_dist": 0,  # these don't affect anything when generating predictions
             "prior_aux_param": 1.0,  # these don't affect anything when generating predictions
-            "sdy": 1.0,
         }
 
+        prepare_X_data(dat, X_clean)
+
+        dat["num_prior_aux_params"] = self.prior_aux_["num_prior_aux_params"]
+        dat["prior_aux_params"] = self.prior_aux_["prior_aux_params"]
+
         # known that fitted with sampling, so fitted_samples is not None
         predicGQ = self.model_.generate_quantities(
             dat,
@@ -708,8 +701,14 @@ class GLM(CoreEstimator):
     def predict(
         self,
         X: ArrayLike,
+        return_std: bool = False,
         show_console: bool = False,
-    ) -> NDArray[np.float64]:
+    ) -> Union[
+        NDArray[Union[np.float64, np.int64]],
+        Tuple[
+            NDArray[Union[np.float64, np.int64]], NDArray[Union[np.float64, np.int64]]
+        ],
+    ]:
         """Compute predictions from supplied data using a fitted model.
             This computes the mean of the predicted distribution,
             given by y_sim in predict_distribution().
@@ -723,13 +722,20 @@ class GLM(CoreEstimator):
         ----------
         X : ArrayLike
             Predictor matrix or array of data to use for prediction.
-        show_console : bool, optional
+            X can be sparse. It will be converted to CSR format if it is not already.
+        return_std : bool, optional, default=False
+            If True, return standard deviation of predictions.
+        show_console : bool, optional, default=False
             Printing output of default CmdStanPy console during Stan operations.
 
         Returns
         -------
         NDArray[np.float64]
             Array of predictions of shape (n_samples, 1) made by fitted model.
+
+        NDArray[np.float64]
+            Standard deviation of predictive distribution points.
+
         """
         check_is_fitted(self)
 
@@ -741,11 +747,21 @@ class GLM(CoreEstimator):
 
         X_clean, _ = self._validate_data(X=X, ensure_X_2d=True)
 
-        return self.predict_distribution(  # type: ignore
+        y_predict_mean = self.predict_distribution(
             X_clean,
             show_console=show_console,
         ).mean(axis=0, dtype=np.float64)
 
+        if return_std:
+            sigmas_squared_data = (np.dot(X_clean, self.sigma_) * X_clean).sum(  # type: ignore
+                axis=1
+            )
+            y_std = np.sqrt(sigmas_squared_data + (1.0 / self.alpha_))
+
+            return y_predict_mean, y_std
+        else:
+            return y_predict_mean  # type: ignore
+
     def score(
         self,
         X: ArrayLike,
@@ -760,6 +776,7 @@ class GLM(CoreEstimator):
         X : ArrayLike
             Matrix or array of predictors having shape (n_samples, n_predictors)
             that consists of test data.
+            X can be sparse. It will be converted to CSR format if it is not already.
         y : ArrayLike
             Array of shape (n_samples,) containing the target values
             corresponding to given test dat X.
@@ -787,23 +804,26 @@ class GLM(CoreEstimator):
 
         return 1 - ssreg / sstot
 
-    def _more_tags(self) -> Dict[str, Any]:
-        """
-        Sets tags for current model that exclude certain sk-learn estimator
-        checks that are not applicable to this model.
-        """
-        return {
-            "_xfail_checks": {
-                "check_methods_sample_order_invariance": "check is not applicable.",
-                "check_methods_subset_invariance": "check is not applicable.",
-                "check_fit_idempotent": """model is idempotent, but not to the required degree of
-                    accuracy as this is a probabilistic setting.""",
-                "check_fit1d": """provided automatic cast from 1d to 2d in data validation.""",
-                # NOTE: the expected behavior here is to raise a ValueError, the package intends
-                # to give alternative default behavior in these scenarios!
-                "check_fit2d_predict1d": """provided automatic cast from 1d to 2d in data validation
-                 STILL NEEDS TO BE INVESTIGATED FOR GQ ISSUE""",
-                # NOTE: the expected behavior here is to raise a ValueError,
-                #  the package intends to give alternative default behavior in these scenarios!
-            }
-        }
+
+def prepare_X_data(
+    data: Dict[str, Any],
+    X: Union[NDArray[np.int64], NDArray[np.float64], sp.csr_matrix],
+) -> None:
+    """
+    Populate data dictionary with the (possibly sparse) X matrix.
+    """
+    if sp.issparse(X):
+        assert isinstance(X, sp.csr_matrix)
+        data["X_dense"] = 0
+        data["X"] = []
+        data["X_data"] = X.data
+        data["X_nz"] = X.nnz
+        data["X_idxs"] = X.indices + 1
+        data["X_indptr"] = X.indptr + 1
+    else:
+        data["X_dense"] = 1
+        data["X"] = X
+        data["X_nz"] = 0
+        data["X_data"] = []
+        data["X_idxs"] = []
+        data["X_indptr"] = []
