@@ -2,12 +2,72 @@
 
 import numpy as np
 import pytest
-import scipy.stats as stats  # type: ignore
+import scipy.sparse as sp
+import scipy.stats as stats
 from data import _gen_fam_dat_continuous, _gen_fam_dat_discrete, bcdata_dict
 from sklearn.utils.estimator_checks import check_estimator  # type: ignore
 
 from scikit_stan.generalized_linear_regression import GLM
-from scikit_stan.modelcore import CoreEstimator
+
+
+def test_sparse() -> None:
+    """Test sparse X"""
+
+    glm = GLM(family="gaussian", seed=1234)
+
+    gaussian_dat_X, gaussian_dat_Y = _gen_fam_dat_continuous(
+        family="gaussian", link="identity"
+    )
+
+    glm.fit(X=sp.csr_matrix(gaussian_dat_X[:, np.newaxis]), y=gaussian_dat_Y)
+
+    reg_coeffs = np.array([])
+    for val in [glm.alpha_, glm.beta_]:
+        reg_coeffs = np.append(reg_coeffs, val)
+
+    np.testing.assert_allclose(reg_coeffs, np.array([0.6, 0.2]), rtol=1e-1, atol=1e-1)
+
+    glm = GLM(family="poisson", link="log", seed=1234)
+
+    poisson_dat_X, poisson_dat_Y = _gen_fam_dat_discrete(family="poisson", link="log")
+
+    fitted = glm.fit(X=poisson_dat_X, y=poisson_dat_Y)
+
+    assert (
+        fitted.fitted_samples_.summary()["5%"]["alpha[1]"]
+        <= 0.6
+        <= fitted.fitted_samples_.summary()["95%"]["alpha[1]"]
+    )
+    assert (
+        fitted.fitted_samples_.summary()["5%"]["beta[1]"]
+        <= 0.2
+        <= fitted.fitted_samples_.summary()["95%"]["beta[1]"]
+    )
+
+
+def test_no_intercept_regression() -> None:
+    """Ensure that no intercept is added if fit_intercept is False
+    and that the default behavior is fit_intercept=True."""
+    X, y = _gen_fam_dat_continuous(family="gaussian", link="log", seed=1234321)
+
+    glm_no_intercept = GLM(
+        family="gaussian", link="log", fit_intercept=False, seed=1234
+    )
+    glm_intercept = GLM(family="gaussian", link="log", fit_intercept=True, seed=1234)
+    glm_no_intercept.fit(X, y)
+    glm_intercept.fit(X, y)
+
+    assert "alpha[1]" not in list(
+        glm_no_intercept.fitted_samples_.summary().head().index
+    )
+    assert "alpha[1]" in list(glm_intercept.fitted_samples_.summary().head().index)
+
+
+@pytest.mark.slow
+# @pytest.mark.parametrize("estimator", [GLM()])
+def test_compatible_estimator() -> None:
+    """Ensure that GLM Estimator is sk-learn compatible."""
+    check_estimator(GLM())
 
 
 @pytest.mark.parametrize(
@@ -40,12 +100,6 @@ def test_GLM_alg_params_correct(alg_args) -> None:
     glm.fit(X=gamma_dat_X, y=gamma_dat_Y)
 
 
-@pytest.mark.slow
-@pytest.mark.parametrize("estimator", [GLM()])
-def test_compatible_estimator(estimator: "CoreEstimator") -> None:
-    check_estimator(estimator)
-
-
 def test_notfittederror_glm() -> None:
     glm = GLM(seed=1234)
     with pytest.raises(Exception) as e_info:
@@ -76,9 +130,9 @@ def test_prior_config_default_nongaussian(prior_config) -> None:
     }
 
     assert (
-        fitted.fitted_samples_.summary()["5%"]["alpha"] - 0.01
+        fitted.fitted_samples_.summary()["5%"]["alpha[1]"] - 0.01
         <= 0.6
-        <= fitted.fitted_samples_.summary()["95%"]["alpha"]
+        <= fitted.fitted_samples_.summary()["95%"]["alpha[1]"]
     )
 
 
@@ -101,9 +155,9 @@ def test_prior_config_default_gaussian(prior_config) -> None:
     }
 
     assert (
-        fitted.fitted_samples_.summary()["5%"]["alpha"] - 0.01
+        fitted.fitted_samples_.summary()["5%"]["alpha[1]"] - 0.01
         <= 0.6
-        <= fitted.fitted_samples_.summary()["95%"]["alpha"]
+        <= fitted.fitted_samples_.summary()["95%"]["alpha[1]"]
     )
 
 
@@ -344,7 +398,7 @@ def test_gaussian_link_scipy_gen(link: str):
     glm = GLM(family="gaussian", link=link, seed=1234)
 
     gaussian_dat_X, gaussian_dat_Y = _gen_fam_dat_continuous(
-        family="gaussian", link=link
+        family="gaussian", link=link, Nsize=1000
     )
 
     glm.fit(X=gaussian_dat_X, y=gaussian_dat_Y)
@@ -355,21 +409,27 @@ def test_gaussian_link_scipy_gen(link: str):
 
     np.testing.assert_allclose(reg_coeffs, np.array([0.6, 0.2]), rtol=1e-1, atol=1e-1)
 
+    # very rough posterior predictive
+    out_y = glm.predict(gaussian_dat_X)
+    np.testing.assert_allclose(
+        gaussian_dat_Y.mean(), out_y.mean(), rtol=1e-1, atol=1e-1
+    )
+
 
 @pytest.mark.parametrize("link", ["identity", "log", "inverse"])
 def test_gamma_link_scipy_gen(link: str) -> None:
-    glm = GLM(family="gamma", link=link, seed=1234)
+    glm = GLM(family="gamma", link=link, seed=1234, algorithm_params={})
 
     gamma_dat_X, gamma_dat_Y = _gen_fam_dat_continuous(
-        family="gamma", link=link, Nsize=100
+        family="gamma", link=link, Nsize=1000
     )
 
     fitted = glm.fit(X=gamma_dat_X, y=gamma_dat_Y)
 
     assert (
-        fitted.fitted_samples_.summary()["5%"]["alpha"] - 0.01
+        fitted.fitted_samples_.summary()["5%"]["alpha[1]"] - 0.01
         <= 0.6
-        <= fitted.fitted_samples_.summary()["95%"]["alpha"]
+        <= fitted.fitted_samples_.summary()["95%"]["alpha[1]"]
     )
     assert (
         fitted.fitted_samples_.summary()["5%"]["beta[1]"] - 0.1
@@ -377,6 +437,10 @@ def test_gamma_link_scipy_gen(link: str) -> None:
         <= fitted.fitted_samples_.summary()["95%"]["beta[1]"] + 0.1
     )
 
+    # very rough posterior predictive
+    out_y = glm.predict(gamma_dat_X)
+    np.testing.assert_allclose(gamma_dat_Y.mean(), out_y.mean(), rtol=1e-1, atol=1e-1)
+
 
 @pytest.mark.parametrize("lotnumber", ["lot1", "lot2"])
 def test_gamma_bloodclotting(lotnumber: str) -> None:
@@ -418,9 +482,9 @@ def test_invgaussian_link_scipy_gen(link: str):
     fitted = glm.fit(X=invgaussian_dat_X, y=invgaussian_dat_Y)
 
     assert (
-        fitted.fitted_samples_.summary()["5%"]["alpha"] - 0.02
+        fitted.fitted_samples_.summary()["5%"]["alpha[1]"] - 0.02
         <= 0.6
-        <= fitted.fitted_samples_.summary()["95%"]["alpha"] + 0.02
+        <= fitted.fitted_samples_.summary()["95%"]["alpha[1]"] + 0.02
     )
     assert (
         fitted.fitted_samples_.summary()["5%"]["beta[1]"] - 0.02
@@ -434,8 +498,8 @@ def test_invgaussian_link_scipy_gen(link: str):
     [
         None,
         {},
-        {"prior_aux_dist": "exponential", "prior_aux_param": 0.5},
-        {"prior_aux_dist": "chi2", "prior_aux_param": 2.5},
+        {"prior_aux_dist": "exponential", "beta": 0.5},
+        {"prior_aux_dist": "chi2", "nu": 2.5},
     ],
 )
 def test_glm_prior_aux_setup(prior_aux) -> None:
@@ -450,14 +514,30 @@ def test_glm_prior_aux_setup(prior_aux) -> None:
 
     if prior_aux is None:
         """Default unscaled prior."""
-        assert glm.prior_aux_ == {"prior_aux_dist": 0, "prior_aux_param": 1.0}
+        assert glm.prior_aux_ == {
+            "num_prior_aux_params": 1,
+            "prior_aux_dist": 0,
+            "prior_aux_params": [1.0],
+        }
     elif len(prior_aux) == 0:
-        assert glm.prior_aux_ == {"prior_aux_dist": -1, "prior_aux_param": 0.0}
+        assert glm.prior_aux_ == {
+            "num_prior_aux_params": 1,
+            "prior_aux_dist": -1,
+            "prior_aux_params": [0.0],
+        }
     else:
         if prior_aux["prior_aux_dist"] == "exponential":
-            assert glm.prior_aux_ == {"prior_aux_dist": 0, "prior_aux_param": 0.5}
+            assert glm.prior_aux_ == {
+                "num_prior_aux_params": 1,
+                "prior_aux_dist": 0,
+                "prior_aux_params": [0.5],
+            }
         else:
-            assert glm.prior_aux_ == {"prior_aux_dist": 1, "prior_aux_param": 2.5}
+            assert glm.prior_aux_ == {
+                "num_prior_aux_params": 1,
+                "prior_aux_dist": 1,
+                "prior_aux_params": [2.5],
+            }
 
 
 # NOTE: for the identity link, the generated data may lead to a negative lambda
@@ -483,9 +563,9 @@ def test_poisson_link_scipy_gen(link: str):
     fitted = glm.fit(X=poisson_dat_X, y=poisson_dat_Y)
 
     assert (
-        fitted.fitted_samples_.summary()["5%"]["alpha"]
+        fitted.fitted_samples_.summary()["5%"]["alpha[1]"]
         <= 0.6
-        <= fitted.fitted_samples_.summary()["95%"]["alpha"]
+        <= fitted.fitted_samples_.summary()["95%"]["alpha[1]"]
     )
     assert (
         fitted.fitted_samples_.summary()["5%"]["beta[1]"]
@@ -511,9 +591,9 @@ def test_poisson_sklearn_poissonregressor():
         reg_coeffs = np.append(reg_coeffs, val)
 
     assert (
-        fitted.fitted_samples_.summary()["5%"]["alpha"]
+        fitted.fitted_samples_.summary()["5%"]["alpha[1]"]
         <= 2.088
-        <= fitted.fitted_samples_.summary()["95%"]["alpha"]
+        <= fitted.fitted_samples_.summary()["95%"]["alpha[1]"]
     )
     assert (
         fitted.fitted_samples_.summary()["5%"]["beta[1]"]
@@ -527,139 +607,26 @@ def test_poisson_sklearn_poissonregressor():
     )
 
 
-@pytest.mark.skip(reason="Inconsistency with R regression; TBI")
-# TODO: this should have 4 regression coefficients?
 def test_poisson_rstanarm_data():
     # NOTE: this data comes from rstanarm tests
     X = np.array(
-        [[1, 1], [1, 2], [1, 3], [2, 1], [2, 2], [2, 3], [3, 1], [3, 2], [3, 3]]
+        [
+            [1, 1, 1],
+            [1, 2, 1],
+            [1, 3, 1],
+            [1, 1, 2],
+            [1, 2, 2],
+            [1, 3, 2],
+            [1, 1, 3],
+            [1, 2, 3],
+            [1, 3, 3],
+        ]
     )
 
-    y = [18, 17, 15, 20, 10, 20, 25, 13, 12]
+    y = np.array([18, 17, 15, 20, 10, 20, 25, 13, 12])
 
-    glm_poisson = GLM(family="poisson", link="log", seed=1234, algorithm="MLE")
+    glm_poisson = GLM(family="poisson", link="log", seed=1234, algorithm="optimize")
 
     glm_poisson.fit(X=X, y=y)
 
     assert True
-
-
-if __name__ == "__main__":
-    # from scipy.special import expit  # type: ignore
-    # import matplotlib.pyplot as plt
-
-    # from data import bcdata_dict
-    # NOTE: rate parameter sometimes becomes negative for poisson?
-    # glm = GLM(family="bernoulli")
-    # glm = GLM(family="gamma", link="inverse")
-    # glm = GLM(family="poisson", link="log", algorithm="MLE")
-
-    # X, y = _gen_fam_dat_discrete(
-    #    "poisson",
-    #    alpha=0.9,
-    #    beta=0.2,
-    #    sample_size=30,
-    #    poisson_lambda=1.0,
-    #    link="a",
-    #    num_yn=10,
-    # )
-    # print(X.shape, y.shape)
-    # X = [[1, 2], [2, 3], [3, 4], [4, 3]]
-    # y = [12, 17, 22, 21]
-    # glm.fit(X, y)
-    #
-    # print(glm.alpha_, glm.beta_)
-
-    # X = np.array([
-    #            [1,       1],
-    #            [1,       2],
-    #            [1,       3],
-    #            [2,       1],
-    #            [2,       2],
-    #            [2,       3],
-    #            [3,       1],
-    #            [3,       2],
-    #            [3,       3]])
-    #
-    # y = [18, 17, 15, 20, 10, 20, 25, 13, 12]
-    # from scipy import stats
-
-    rng = np.random.default_rng(seed=1234)
-    glm = GLM(family="inverse-gaussian", link="identity", seed=1234)
-    # beta = stats.norm.rvs(5, 1, size=1)
-    # alpha = stats.norm.rvs(5, 1, size=1)
-
-    X = stats.norm.rvs(0, 1, size=(100,))
-    y = rng.wald(0.6 + 0.2 * X, 0.3)
-
-    # y = rng.poisson(0.6 + 0.2 * np.exp(X))
-    # y = stats.invgauss.rvs(0.6 + 0.2 * np.exp(X))
-
-    # y = rng.poisson(0.6 + 0.2 * X)
-    # y = rng.normal(1 / (0.6 + X * 0.2))
-    # y = rng.gamma(1 / (0.6 + X * 0.2))
-    # _gen_fam_dat_continuous(family="gamma", link=link)
-    fit = glm.fit(X=X, y=y)
-
-    print(fit.fitted_samples_.summary())
-    print(
-        fit.fitted_samples_.summary()["5%"]["alpha"],
-        fit.fitted_samples_.summary()["95%"]["alpha"],
-    )
-    print(
-        glm.fitted_samples_.summary()["5%"]["beta[1]"],
-        glm.fitted_samples_.summary()["95%"]["beta[1]"],
-    )
-    # print(beta, alpha)
-
-    # for link in ["identity", "log", "inverse"]:
-    #    glm = GLM(family="gamma", link=link, seed=1234)
-#
-#    gamma_dat_X, gamma_dat_Y = _gen_fam_dat_continuous(family="gamma", link=link)
-#
-#    glm.fit(X=gamma_dat_X, y=gamma_dat_Y)
-#
-#    reg_coeffs = np.array([])
-#    for val in [glm.alpha_, glm.beta_]:
-#        reg_coeffs = np.append(reg_coeffs, val)
-#
-#
-# X, y = _gen_fam_dat_continuous(family="gamma", link="log")
-# glm_poisson = GLM(family="gamma", link="log", algorithm="MLE")
-#
-# glm_poisson.fit(X=X, y=y)
-#
-# print(glm_poisson.alpha_, glm_poisson.beta_)
-
-# binom_data_X, binom_data_y = _gen_fam_dat_discrete(
-#    "binomial", "a", 0.7, np.array([0.4]), 20, 30
-# )
-# glm.fit(X=binom_data_X, y=binom_data_y, show_console=True)
-# gamma_dat_X, gamma_dat_Y = _gen_fam_dat(
-#    "inverse-gaussian", Nsize=1000, alpha=0.9, beta=0.3, mu=0.7, sigma=1.9
-# )
-# gauss_dat_X, gauss_dat_y = _gen_fam_dat(
-#    "gaussian", Nsize=1000, alpha=0.9, beta=0.3
-# )
-# bc_data_y, bc_data_X = np.log(bcdata_dict["u"]), np.column_stack(
-#    (bcdata_dict["lot1"], bcdata_dict["lot2"])
-# )
-# bc_data_X, bc_data_y = np.log(bcdata_dict["u"]), bcdata_dict["lot2"]
-# glm.fit(X=bc_data_X, y=bc_data_y, show_console=True)
-# glm.fit(X=gamma_dat_X, y=gamma_dat_Y, show_console=False)
-# print(glm.alpha_, glm.beta_)
-# glm.fit(X=gauss_dat_X, y=gauss_dat_y, show_console=True)
-# glm.fit(X=bc_data_X, y=bc_data_y, show_console=True)
-# predics = glm.predict(X=gauss_dat_X)
-# predics = glm.predict(X=gamma_dat_X)
-# plt.scatter(gauss_dat_X, gauss_dat_y)
-# plt.scatter(gauss_dat_X, predics)
-# plt.hist(gamma_dat_Y, density=True, histtype="stepfilled", alpha=0.2)
-# plt.hist(predics, density=True, histtype="stepfilled", alpha=0.2)
-# plt.scatter(gamma_dat_X, gamma_dat_Y)
-# plt.scatter(gamma_dat_X, predics)
-
-# plt.show()
-# print(glm.predict(X=bc_data_X, show_console=False))
-# print(glm.fit(X=xdat, y=ydat, show_console=True))
-# print(glm.predict(X=xdat, show_console=True))
